# Copyright (c) 2021 J. Lewis Muir
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

require 'fileutils'
require 'optparse'

require_relative 'application'

#
# Rewrites CBF images.
#
class CBFRewriteApp < Application
  class Error < StandardError
  end

  HeaderField = Struct.new(:name, :value_start)
  IOPair = Struct.new(:in, :out)
  Options = Struct.new(:help, :preserve, :verbose, :header)

  # The maximum header line length is 2048 characters, excluding
  # line termination, which is 2048 bytes since the character
  # set is US-ASCII.  See the CIF 1.1 specification at
  # <https://www.iucr.org/resources/cif/spec/version1.1/cifsyntax>.
  MAX_LINE_LENGTH_IN_BYTES = 2048 + 2

  TIMESTAMP_PHEADER_FIELD_NAME = 'Pseudo.Timestamp'

  # The DECTRIS CBF Header Specification says that the left and right
  # parenthesis characters (0x28 and 0x29, respectively) are considered
  # whitespace.  This implementation deviates from that part of the spec:
  # they are not considered whitespace.  This is because when treating the
  # header as a series of field name-value pairs (to make it easy to specify
  # a new field value), the value could end up containing an unbalanced right
  # parenthesis which would be awkward.  For example, in accordance with
  # the spec, considering the left and right parenthesis characters to be
  # whitespace in
  #
  #   # Beam_xy (1570.00, 1676.50) pixels\r\n
  #
  # would result in a "Beam_xy" field value of
  #
  #   1570.00, 1676.50) pixels
  #
  # By not considering the left and right parenthesis characters to be
  # whitespace, the resulting value includes the left parenthesis
  #
  #   (1570.00, 1676.50) pixels
  #
  # which is the more expected behavior.
  HEADER_FIELD_RE = /^# ++([^ #=:,]++)[ #=:,]++/
  SI_SENSOR_HEADER_FIELD_RE = /^# ++(Silicon sensor, thickness)[ #=:,]++/
  HEADER_TIMESTAMP_RE = /^# ++(\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+)\r$/

  # Converts SystemCallError to Error with pathname-first message.
  def self.open_file(pathname, *args, &block)
    File.open(pathname, *args, &block)
  rescue SystemCallError => e
    raise Error.new("#{pathname}: #{SystemCallError.new(e.errno)}")
  end

  # Converts SystemCallError to Error with pathname-first message.
  def self.stat_file(pathname)
    File.stat(pathname)
  rescue SystemCallError => e
    raise Error.new("#{pathname}: #{SystemCallError.new(e.errno)}")
  end

  def self.parse_header_field(line)
    # The "Silicon sensor, thickness" field name is weird because it contains
    # whitespace-equivalent characters, so handle it as a special case.
    m = SI_SENSOR_HEADER_FIELD_RE.match(line)
    return HeaderField.new(m[1], m.end(0)) if m
    # The acquisition timestamp is weird because it has no field name, so
    # handle it as a special case.
    m = HEADER_TIMESTAMP_RE.match(line)
    return HeaderField.new(TIMESTAMP_PHEADER_FIELD_NAME, m.begin(1)) if m
    # Handle the rest of the field names in the normal way.
    m = HEADER_FIELD_RE.match(line)
    m ? HeaderField.new(m[1], m.end(0)) : nil
  end

  def initialize(**args)
    super
    @options = nil
    @exit_status = 0
  end

  def usage_message
    "usage: #{@prog_name} [-hpv] [-H <name>=<value>] [<file> ...]"
  end

  def update_exit_status(status)
    @exit_status = status if @exit_status == 0
  end

  def parse_options
    @options = Options.new
    @options.help = false
    @options.preserve = false
    @options.verbose = false
    @options.header = {}
    begin
      OptionParser.new do |opts|
        opts.on('-h', '--help') { @options.help = true }
        opts.on('-p') { @options.preserve = true }
        opts.on('-v') { @options.verbose = true }
        opts.on('-H PAIR') do |arg|
          pair = arg.split('=', 2)
          if pair.size != 2
            warnx("-H #{arg}: Invalid name-value pair")
            usage
          end
          @options.header[pair.first] = pair.last
        end
      end.parse!(@argv)
    rescue OptionParser::ParseError => e
      warnx(e)
      usage
    end
  end

  def create_rewrite_file_io_pair(pathname)
    f_in = self.class.open_file(pathname, 'rb')
    begin
      f_out = nil
      f_out = self.class.open_file(
        "#{pathname}.tmp",
        File::WRONLY | File::BINARY | File::CREAT | File::EXCL
      )
      IOPair.new(f_in, f_out)
    rescue StandardError => e
      f_in.close
      f_out.close if f_out
      raise
    end
  end

  def rewrite(io_pair)
    # Skip to the beginning of the header_contents section.
    loop do
      line = io_pair.in.gets("\r\n", MAX_LINE_LENGTH_IN_BYTES)
      return if line.nil?
      if !line.end_with?("\r\n")
        pathname = io_pair.in.respond_to?(:path) ? io_pair.in.path : '-'
        warnx("#{pathname}: Header line too long")
        update_exit_status(1)
        io_pair.out.write(line)
        IO.copy_stream(io_pair.in, io_pair.out)
        return
      end
      io_pair.out.write(line)
      break if line == "_array_data.header_contents\r\n"
    end
    # Rewrite header fields.
    loop do
      line = io_pair.in.gets("\r\n", MAX_LINE_LENGTH_IN_BYTES)
      return if line.nil?
      if !line.end_with?("\r\n")
        pathname = io_pair.in.respond_to?(:path) ? io_pair.in.path : '-'
        warnx("#{pathname}: Header line too long")
        update_exit_status(1)
        io_pair.out.write(line)
        IO.copy_stream(io_pair.in, io_pair.out)
        return
      end
      if line == "_array_data.data\r\n"
        io_pair.out.write(line)
        break
      end
      if line == ";\r\n"
        io_pair.out.write(line)
        next
      end
      field = self.class.parse_header_field(line)
      if !field
        io_pair.out.write(line)
        next
      end
      new_value = @options.header[field.name]
      if !new_value
        io_pair.out.write(line)
        next
      end
      io_pair.out.write("#{line[0..(field.value_start-1)]}#{new_value}\r\n")
    end
    # Write the rest of the file unmodified.
    IO.copy_stream(io_pair.in, io_pair.out)
  end

  # Sets the user ID, group ID, file mode, access time, and modification
  # time from the source file stat to the destination file at the specified
  # pathname, as allowed by permissions.
  #
  # See also:
  # * <https://man.netbsd.org/cp.1>
  # * <https://man.netbsd.org/chmod.2>
  # * <https://nxr.netbsd.org/xref/src/bin/cp/utils.c#setfile>
  def set_file_attributes(fs_src, fp_dst)
    # Keep only S_ISUID | S_ISGID | S_IRWXU | S_IRWXG | S_IRWXO.
    mode = fs_src.mode & 06777
    begin
      File.chown(fs_src.uid, fs_src.gid, fp_dst)
    rescue SystemCallError => e
      if !defined?(Errno::EPERM) || !e.is_a?(Errno::EPERM)
        warnx("#{fp_dst}: chown: #{SystemCallError.new(e.errno)}")
        update_exit_status(1)
      end
      # Drop S_ISUID | S_ISGID.
      mode &= ~06000
    end
    begin
      File.chmod(mode, fp_dst)
    rescue SystemCallError => e
      warnx("#{fp_dst}: chmod: #{SystemCallError.new(e.errno)}")
      update_exit_status(1)
    end
    begin
      File.utime(fs_src.atime, fs_src.mtime, fp_dst)
    rescue SystemCallError => e
      warnx("#{fp_dst}: utime: #{SystemCallError.new(e.errno)}")
      update_exit_status(1)
    end
  end

  def run
    parse_options
    usage if @options.help
    @argv << '-' if @argv.empty?
    @argv.each do |pathname|
      begin
        if pathname == '-'
          @stdin.binmode
          @stdout.binmode
          rewrite(IOPair.new(@stdin, @stdout))
        else
          iop = create_rewrite_file_io_pair(pathname)
          begin
            @stdout.puts(pathname) if @options.verbose
            rewrite(iop)
            iop.in.close
            iop.out.close
            if @options.preserve
              set_file_attributes(
                self.class.stat_file(iop.in.path),
                iop.out.path
              )
            end
            FileUtils.mv(iop.out.path, iop.in.path, force: true)
          ensure
            iop.in.close unless iop.in.closed?
            iop.out.close unless iop.out.closed?
            FileUtils.rm_f(iop.out.path)
          end
        end
      rescue Error, SystemCallError => e
        warnx(e.to_s)
        update_exit_status(1)
      end
    end
    @exit_status
  end
end

exit(CBFRewriteApp.new.run) if __FILE__ == $0
