# Copyright (c) 2021 J. Lewis Muir
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

require 'fileutils'
require 'time'
require 'tmpdir'
require 'stringio'

require 'test/unit'

require_relative 'backport_stringio_binmode'

require_relative 'cbf_rewrite_app'

#
# Tests CBFRewriteApp.
#
class CBFRewriteAppTest < Test::Unit::TestCase
  CBF_IMAGE_FRAGMENT =
    "###CBF: VERSION 1.5, CBFlib v0.7.8 - PILATUS detectors\r\n" \
    "\r\n" \
    "data_Image_000004\r\n" \
    "\r\n" \
    "_array_data.header_convention \"PILATUS_1.2\"\r\n" \
    "_array_data.header_contents\r\n" \
    ";\r\n" \
    "# Detector: Dectris Eiger2 9M, S/N E-18-0110\r\n" \
    "# 2020-02-06T02:39:12.602\r\n" \
    "# Pixel_size 75e-6 m x 75e-6 m\r\n" \
    "# Silicon sensor, thickness 0.000450 m\r\n" \
    "# Exposure_time 0.2499999 s\r\n" \
    "# Exposure_period 0.2500000 s\r\n" \
    "# Count_cutoff 1102870 counts\r\n" \
    "# Threshold_setting: 6199.196303 eV\r\n" \
    "# N_excluded_pixels = 0\r\n" \
    "# Image_path: /\r\n" \
    "# Wavelength 1.00000 A\r\n" \
    "# Detector_distance 0.35000 m\r\n" \
    "# Beam_xy (1570.00, 1676.50) pixels\r\n" \
    "# Start_angle 0.7500 deg.\r\n" \
    "# Angle_increment 0.2500 deg.\r\n" \
    "# Detector_2theta 0.0001 deg.\r\n" \
    "# Filter_transmission 5.8 % \r\n" \
    ";\r\n" \
    "\r\n" \
    "_array_data.data\r\n" \
    ";\r\n" \
    "--CIF-BINARY-FORMAT-SECTION--\r\n" \
    "Content-Type: application/octet-stream;\r\n" \
    "     conversions=\"x-CBF_BYTE_OFFSET\"\r\n" \
    "Content-Transfer-Encoding: BINARY\r\n" \
    "X-Binary-Size: 10138328\r\n" \
    "X-Binary-ID: 1\r\n" \
    "X-Binary-Element-Type: \"signed 32-bit integer\"\r\n" \
    "X-Binary-Element-Byte-Order: LITTLE_ENDIAN\r\n" \
    "Content-MD5: KoJs0BrXhOF/lL/3gvpqVw==\r\n" \
    "X-Binary-Number-of-Elements: 10138296\r\n" \
    "X-Binary-Size-Fastest-Dimension: 3108\r\n" \
    "X-Binary-Size-Second-Dimension: 3262\r\n" \
    "X-Binary-Size-Padding: 0\r\n" \
    "\r\n" \
    "\x0c\x1a\x04\xd5\x00\x00\x00\x00\x00\x00\x00\x00\x00".tap do |x|
      x.force_encoding(Encoding::ASCII_8BIT)
      x.freeze
    end

  @@unix_like_chmod_mutex = Mutex.new
  @@unix_like_chmod = nil

  def self.create_app_context
    {
      prog_name: nil,
      argv: [],
      stdin: StringIO.new('', 'r'),
      stdout: StringIO.new('', 'w'),
      stderr: StringIO.new('', 'w'),
      env: {}
    }
  end

  def self.unix_like_chmod?
    @@unix_like_chmod_mutex.synchronize do
      return @@unix_like_chmod unless @@unix_like_chmod.nil?
      @@unix_like_chmod = Dir.mktmpdir(File.basename($0)) do |dir|
        f_pathname = File.join(dir, 'foo')
        File.open(f_pathname, 'w') { |f| }
        File.chmod(0644, f_pathname)
        File.chmod(0640, f_pathname)
        File.stat(f_pathname).mode & 06777 == 0640
      end
    end
  end

  def test_help_option
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['-h']
    e = assert_raise(SystemExit) { CBFRewriteApp.new(**app_ctx).run }
    assert_equal(1, e.status)
    assert_match(/^usage: /, app_ctx[:stderr].string)
  end

  def test_invalid_header_field_option
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['-H', 'foo']
    e = assert_raise(SystemExit) { CBFRewriteApp.new(**app_ctx).run }
    assert_equal(1, e.status)
    assert_match(/Invalid name-value pair/, app_ctx[:stderr].string)
  end

  def test_stdio_rewrite_sets_binmode
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['-H', 'Exposure_time=0.5 s']
    app_ctx[:stdin] = StringIO.new(CBF_IMAGE_FRAGMENT.dup, 'r')
    status = CBFRewriteApp.new(**app_ctx).run
    assert_equal(0, status)
    assert_equal(Encoding::ASCII_8BIT, app_ctx[:stdin].external_encoding)
    assert_equal(Encoding::ASCII_8BIT, app_ctx[:stdout].external_encoding)
    actual = app_ctx[:stdout].string
    assert_match(/^# Exposure_time 0.5 s\r$/, actual)
  end

  def test_stdio_zero_length
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['-H', 'Exposure_time=0.5 s']
    status = CBFRewriteApp.new(**app_ctx).run
    assert_equal(0, status)
    assert_empty(app_ctx[:stdout].string)
  end

  def test_stdio_rewrite_exposure_time
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['-H', 'Exposure_time=0.5 s']
    app_ctx[:stdin] = StringIO.new(CBF_IMAGE_FRAGMENT.dup, 'r')
    status = CBFRewriteApp.new(**app_ctx).run
    assert_equal(0, status)
    actual = app_ctx[:stdout].string
    assert_match(/^# Exposure_time 0.5 s\r$/, actual)
  end

  def test_stdio_rewrite_timestamp
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['-H', 'Pseudo.Timestamp=2021-01-02T03:45:01.234']
    app_ctx[:stdin] = StringIO.new(CBF_IMAGE_FRAGMENT.dup, 'r')
    status = CBFRewriteApp.new(**app_ctx).run
    assert_equal(0, status)
    actual = app_ctx[:stdout].string
    assert_match(/^# 2021-01-02T03:45:01\.234\r$/, actual)
  end

  def test_stdio_rewrite_all
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = [
      '-H', 'Detector=Acme Ultra 9M, S/N 12345',
      '-H', 'Pseudo.Timestamp=2021-01-02T03:45:01.234',
      '-H', 'Pixel_size=74e-6 m x 74e-6 m',
      '-H', 'Silicon sensor, thickness=0.000350 m',
      '-H', 'Exposure_time=0.4999999 s',
      '-H', 'Exposure_period=0.5000000 s',
      '-H', 'Count_cutoff=1000000 counts',
      '-H', 'Threshold_setting=6180.0 eV',
      '-H', 'N_excluded_pixels=1',
      '-H', 'Image_path=/foo',
      '-H', 'Wavelength=1.10000 A',
      '-H', 'Detector_distance=0.25000 m',
      '-H', 'Beam_xy=(1571.00, 1677.50) pixels',
      '-H', 'Start_angle=0.8500 deg.',
      '-H', 'Angle_increment=0.1500 deg.',
      '-H', 'Detector_2theta=0.0002 deg.',
      '-H', 'Filter_transmission=6.0 % '
    ]
    app_ctx[:stdin] = StringIO.new(CBF_IMAGE_FRAGMENT.dup, 'r')
    status = CBFRewriteApp.new(**app_ctx).run
    assert_equal(0, status)
    actual = app_ctx[:stdout].string
    assert_match(/^# Detector: Acme Ultra 9M, S\/N 12345\r$/, actual)
    assert_match(/^# 2021-01-02T03:45:01\.234\r$/, actual)
    assert_match(/^# Pixel_size 74e-6 m x 74e-6 m\r$/, actual)
    assert_match(/^# Silicon sensor, thickness 0\.000350 m\r$/, actual)
    assert_match(/^# Exposure_time 0\.4999999 s\r$/, actual)
    assert_match(/^# Exposure_period 0\.5000000 s\r$/, actual)
    assert_match(/^# Count_cutoff 1000000 counts\r$/, actual)
    assert_match(/^# Threshold_setting: 6180\.0 eV\r$/, actual)
    assert_match(/^# N_excluded_pixels = 1\r$/, actual)
    assert_match(/^# Image_path: \/foo\r$/, actual)
    assert_match(/^# Wavelength 1\.10000 A\r$/, actual)
    assert_match(/^# Detector_distance 0\.25000 m\r$/, actual)
    assert_match(/^# Beam_xy \(1571\.00, 1677\.50\) pixels\r$/, actual)
    assert_match(/^# Start_angle 0\.8500 deg\.\r$/, actual)
    assert_match(/^# Angle_increment 0\.1500 deg\.\r$/, actual)
    assert_match(/^# Detector_2theta 0\.0002 deg\.\r$/, actual)
    assert_match(/^# Filter_transmission 6\.0 % \r$/, actual)
  end

  def test_stdio_header_line_too_long
    det_line_start = '# Detector: '
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['-H', 'Exposure_time=0.5 s']
    app_ctx[:stdin] = StringIO.new(
      CBF_IMAGE_FRAGMENT.dup.sub(
        Regexp.new("^#{Regexp.quote(det_line_start)}.*+"),
        "#{det_line_start}#{'x' * (2048 - det_line_start.size + 1)}\r\n"
      ),
      'r'
    )
    status = CBFRewriteApp.new(**app_ctx).run
    assert_equal(1, status)
    assert_match(/Header line too long/, app_ctx[:stderr].string)
    actual = app_ctx[:stdout].string
    assert_equal(app_ctx[:stdin].string, actual)
  end

  def test_stdio_rewrite_header_line_too_long_before_header_contents_section
    det_line_start = '# Detector: '
    app_ctx = self.class.create_app_context
    app_ctx[:argv] = ['-H', 'Exposure_time=0.5 s']
    app_ctx[:stdin] = StringIO.new(
      "#{'x' * (2048 + 1)}\r\n#{CBF_IMAGE_FRAGMENT}",
      'r'
    )
    status = CBFRewriteApp.new(**app_ctx).run
    assert_equal(1, status)
    assert_match(/Header line too long/, app_ctx[:stderr].string)
    actual = app_ctx[:stdout].string
    assert_equal(app_ctx[:stdin].string, actual)
  end

  def test_nonexistent_file
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f_000001.cbf')
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['-H', 'Exposure_time=0.5 s', f1_pathname]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(1, status)
      assert_match(
        Regexp.new(": #{Regexp.quote(f1_pathname)}: "),
        app_ctx[:stderr].string
      )
    end
  end

  def test_nonreadable_file
    return unless self.class.unix_like_chmod?
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f_000001.cbf')
      FileUtils.touch(f1_pathname)
      File.chmod(0, f1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['-H', 'Exposure_time=0.5 s', f1_pathname]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(1, status)
      assert_match(
        Regexp.new(": #{Regexp.quote(f1_pathname)}: "),
        app_ctx[:stderr].string
      )
    end
  end

  def test_zero_length_file
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f_000001.cbf')
      FileUtils.touch(f1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['-H', 'Exposure_time=0.5 s', f1_pathname]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_equal(0, File.size(f1_pathname))
    end
  end

  def test_file_rewrite_exposure_time
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f_000001.cbf')
      IO.write(f1_pathname, CBF_IMAGE_FRAGMENT, mode: 'wb')
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['-H', 'Exposure_time=0.5 s', f1_pathname]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(0, status)
      actual = IO.read(f1_pathname, mode: 'rb')
      assert_match(/^# Exposure_time 0.5 s\r$/, actual)
    end
  end

  def test_file_rewrite_header_line_too_long
    Dir.mktmpdir(File.basename($0)) do |dir|
      det_line_start = '# Detector: '
      f1_pathname = File.join(dir, 'f_000001.cbf')
      f1_data = CBF_IMAGE_FRAGMENT.dup.sub(
        Regexp.new("^#{Regexp.quote(det_line_start)}.*+"),
        "#{det_line_start}#{'x' * (2048 - det_line_start.size + 1)}\r\n"
      )
      IO.write(f1_pathname, f1_data, mode: 'wb')
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['-H', 'Exposure_time=0.5 s', f1_pathname]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(1, status)
      assert_match(/Header line too long/, app_ctx[:stderr].string)
      actual = IO.read(f1_pathname, mode: 'rb')
      assert_equal(f1_data, actual)
    end
  end

  def test_file_rewrite_header_line_too_long_before_header_contents_section
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f_000001.cbf')
      f1_data = "#{'x' * (2048 + 1)}\r\n#{CBF_IMAGE_FRAGMENT}"
      IO.write(f1_pathname, f1_data, mode: 'wb')
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['-H', 'Exposure_time=0.5 s', f1_pathname]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(1, status)
      assert_match(/Header line too long/, app_ctx[:stderr].string)
      actual = IO.read(f1_pathname, mode: 'rb')
      assert_equal(f1_data, actual)
    end
  end

  def test_two_files_rewrite_exposure_time
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f_000001.cbf')
      f2_pathname = File.join(dir, 'f_000002.cbf')
      IO.write(f1_pathname, CBF_IMAGE_FRAGMENT, mode: 'wb')
      IO.write(f2_pathname, CBF_IMAGE_FRAGMENT, mode: 'wb')
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['-H', 'Exposure_time=0.5 s', f1_pathname, f2_pathname]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(0, status)
      actual = IO.read(f1_pathname, mode: 'rb')
      assert_match(/^# Exposure_time 0.5 s\r$/, actual)
      actual = IO.read(f2_pathname, mode: 'rb')
      assert_match(/^# Exposure_time 0.5 s\r$/, actual)
    end
  end

  def test_nonfatal_missing_file_rewrite_exposure_time
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f_000001.cbf')
      f2_pathname = File.join(dir, 'f_000002.cbf')
      f3_pathname = File.join(dir, 'f_000003.cbf')
      IO.write(f1_pathname, CBF_IMAGE_FRAGMENT, mode: 'wb')
      IO.write(f3_pathname, CBF_IMAGE_FRAGMENT, mode: 'wb')
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = [
        '-H', 'Exposure_time=0.5 s', f1_pathname, f2_pathname, f3_pathname
      ]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(1, status)
      actual = IO.read(f1_pathname, mode: 'rb')
      assert_match(/^# Exposure_time 0.5 s\r$/, actual)
      assert_match(
        Regexp.new(": #{Regexp.quote(f2_pathname)}: "),
        app_ctx[:stderr].string
      )
      actual = IO.read(f3_pathname, mode: 'rb')
      assert_match(/^# Exposure_time 0.5 s\r$/, actual)
    end
  end

  def test_verbose_rewrite
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f_000001.cbf')
      f2_pathname = File.join(dir, 'f_000002.cbf')
      IO.write(f1_pathname, CBF_IMAGE_FRAGMENT, mode: 'wb')
      IO.write(f2_pathname, CBF_IMAGE_FRAGMENT, mode: 'wb')
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = [
        '-v', '-H', 'Exposure_time=0.5 s', f1_pathname, f2_pathname
      ]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(0, status)
      assert_match(
        Regexp.new(Regexp.quote(File.basename(f1_pathname))),
        app_ctx[:stdout].string
      )
      assert_match(
        Regexp.new(Regexp.quote(File.basename(f2_pathname))),
        app_ctx[:stdout].string
      )
    end
  end

  def test_file_preserve_attributes
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f_000001.cbf')
      IO.write(f1_pathname, CBF_IMAGE_FRAGMENT, mode: 'wb')
      orig_fs1 = File.stat(f1_pathname)
      mode = 0604
      File.chmod(mode, f1_pathname)
      gid = nil
      begin
        gid = Process.groups.find { |x| x != orig_fs1.gid }
      rescue NotImplementedError
        # Process.groups is not implemented on some platforms.
      end
      File.chown(nil, gid, f1_pathname) if gid
      atime = Time.parse('2020-01-02T03:04:05.678')
      mtime = Time.parse('2019-02-03T04:05:06.789')
      File.utime(atime, mtime, f1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['-p', '-H', 'Exposure_time=0.5 s', f1_pathname]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(0, status)
      fs1 = File.stat(f1_pathname)
      if self.class.unix_like_chmod?
        assert_equal(mode, fs1.mode & 06777)
      end
      assert_equal(gid, fs1.gid) if gid
      assert_equal(atime, fs1.atime)
      assert_equal(mtime, fs1.mtime)
    end
  end

  def test_file_preserve_attributes_set_uid_on_exec
    return unless self.class.unix_like_chmod?
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f_000001.cbf')
      IO.write(f1_pathname, CBF_IMAGE_FRAGMENT, mode: 'wb')
      mode = 04604
      File.chmod(mode, f1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['-p', '-H', 'Exposure_time=0.5 s', f1_pathname]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(0, status)
      fs1 = File.stat(f1_pathname)
      assert_equal(mode, fs1.mode & 06777)
    end
  end

  def test_file_preserve_attributes_set_gid_on_exec
    return unless self.class.unix_like_chmod?
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f_000001.cbf')
      IO.write(f1_pathname, CBF_IMAGE_FRAGMENT, mode: 'wb')
      mode = 02604
      File.chmod(mode, f1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['-p', '-H', 'Exposure_time=0.5 s', f1_pathname]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(0, status)
      fs1 = File.stat(f1_pathname)
      assert_equal(mode, fs1.mode & 06777)
    end
  end

  def test_file_preserve_attributes_set_uidgid_on_exec
    return unless self.class.unix_like_chmod?
    Dir.mktmpdir(File.basename($0)) do |dir|
      f1_pathname = File.join(dir, 'f_000001.cbf')
      IO.write(f1_pathname, CBF_IMAGE_FRAGMENT, mode: 'wb')
      mode = 06604
      File.chmod(mode, f1_pathname)
      app_ctx = self.class.create_app_context
      app_ctx[:argv] = ['-p', '-H', 'Exposure_time=0.5 s', f1_pathname]
      status = CBFRewriteApp.new(**app_ctx).run
      assert_equal(0, status)
      fs1 = File.stat(f1_pathname)
      assert_equal(mode, fs1.mode & 06777)
    end
  end
end
